import logging
import os
import abc
import copy
import functools
import multiprocessing.pool

import numpy as np
import pandas as pd
from scipy import stats

from lightfm_pandas.utils import instrumentation, array_math
from lightfm_pandas.data import interactions
from lightfm_pandas import evaluation

logger = logging.getLogger(__name__)


class BaseDFRecommender(abc.ABC, instrumentation.LogLongCallsMeta):
    default_model_params = {}
    default_fit_params = {}

    def __init__(self, user_col='userid', item_col='itemid',
                 rating_col='rating', prediction_col='prediction',
                 model_params=None, fit_params=None, **kwargs):
        self._user_col = user_col
        self._item_col = item_col
        self._item_col_simil = item_col + '_source'
        self._rating_col = rating_col
        self._prediction_col = prediction_col
        self.model_params = self.default_model_params.copy()
        self.fit_params = self.default_fit_params.copy()
        self._set_model_params(model_params)
        self._set_model_params(kwargs)
        self._set_fit_params(fit_params)
        # self.train_df = None
        self.model = None

    @staticmethod
    def toggle_mkl_blas_1_thread(state):
        if state:
            os.environ['MKL_NUM_THREADS'] = '1'
            os.environ['OPENBLAS_NUM_THREADS'] = '1'
        else:
            os.environ.pop('MKL_NUM_THREADS', None)
            os.environ.pop('OPENBLAS_NUM_THREADS', None)

    @staticmethod
    def _dict_update(d, u):
        d = d.copy()
        if u:
            d.update(u)
        return d

    @staticmethod
    def _pop_set_dict(d, params, pop_params):
        for p in pop_params:
            if p in params:
                d[p] = params.pop(p)
        return params

    def _pop_set_params(self, params, pop_params):
        return self._pop_set_dict(self.__dict__, params, pop_params)

    def _set_model_params(self, params):
        self.model_params = self._dict_update(self.model_params, params)

    def _set_fit_params(self, params):
        self.fit_params = self._dict_update(self.fit_params, params)

    def set_params(self, **params):
        """
        this is for skopt / sklearn compatibility
        """
        # pop-set fit_params if provided in bulk
        self._set_fit_params(params.pop('fit_params', {}))
        # pop-set model_params if provided in bulk
        self._set_model_params(params.pop('model_params', {}))
        # pop-set fit_params by keys from default_fit_params if provided flat
        params = self._pop_set_dict(self.fit_params, params, self.default_fit_params.keys())
        # the rest are assumed to be model_params provided flat
        self._set_model_params(params)

    @abc.abstractmethod
    def fit(self, train_obs: interactions.ObservationsDF, *args, **kwargs):
        return self

    @abc.abstractmethod
    def get_recommendations(
            self, user_ids=None, item_ids=None, n_rec=10,
            exclusions=True,
            results_format='lists',
            **kwargs):
        return pd.DataFrame()

    @abc.abstractmethod
    def eval_on_test_by_ranking(
            self, test_dfs, test_names=('',), prefix='rec ',
            include_train=True, items_filter=None, k=10,
            **kwargs):
        return pd.DataFrame()

    @staticmethod
    def _flat_df_to_lists(df, sort_col, group_col, n_cutoff, target_columns):
        order = [group_col] + target_columns
        return df[order]. \
            sort_values(sort_col, ascending=False). \
            groupby(group_col). \
            aggregate(lambda x: list(x)[:n_cutoff]). \
            reset_index()

    def _recos_flat_to_lists(self, df, n_cutoff=None):
        return self._flat_df_to_lists(
            df,
            n_cutoff=n_cutoff,
            sort_col=self._prediction_col,
            group_col=self._user_col,
            target_columns=[self._item_col, self._prediction_col])

    def _simil_flat_to_lists(self, df, n_cutoff=None):
        return self._flat_df_to_lists(
            df,
            n_cutoff=n_cutoff,
            sort_col=self._prediction_col,
            group_col=self._item_col_simil,
            target_columns=[self._item_col, self._prediction_col])

    def _format_results_df(self, source_vec, results_format,
                           target_ids_mat=None, scores_mat=None,
                           target_ids_lists=None, scores_lists=None):

        matrices = target_ids_mat is not None and scores_mat is not None
        lists = target_ids_lists is not None and scores_lists is not None
        assert matrices or lists, 'Provide either matrices or lists'

        if 'recommendations' in results_format:
            source_col = self._user_col
            target_col = self._item_col
            scores_col = self._prediction_col
        elif 'similarities' in results_format:
            source_col = self._item_col_simil
            target_col = self._item_col
            scores_col = self._prediction_col
        else:
            raise NotImplementedError('results_format: ' + results_format)

        order = [source_col, target_col, scores_col]

        if 'lists' in results_format:
            if matrices:
                return pd.DataFrame({
                    source_col: source_vec,
                    target_col: list(target_ids_mat),
                    scores_col: list(scores_mat)
                })[order]
            else:
                return pd.DataFrame({
                    source_col: source_vec,
                    target_col: target_ids_lists,
                    scores_col: scores_lists
                })[order]

        elif 'flat' in results_format:
            if matrices:
                n_rec = target_ids_mat.shape[1]
                return pd.DataFrame({
                    source_col: np.array(source_vec).repeat(n_rec),
                    target_col: np.concatenate(list(target_ids_mat) if len(target_ids_mat) else [[]]),
                    scores_col: np.concatenate(list(scores_mat) if len(target_ids_mat) else [[]]),
                })[order]
            else:
                return pd.DataFrame({
                    source_col: np.array(source_vec).repeat(list(map(len, target_ids_lists))),
                    target_col: np.concatenate(target_ids_lists if len(target_ids_lists) else [[]]),
                    scores_col: np.concatenate(scores_lists if len(scores_lists) else [[]]),
                })[order]

        else:
            raise NotImplementedError('results_format: ' + results_format)


class BaseDFSparseRecommender(BaseDFRecommender):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.sparse_mat_builder = None
        self.train_mat = None
        self.exclude_mat = None
        self.external_features_mat = None
        self.items_handler = None

    def user_inds(self, user_ids):
        return self.sparse_mat_builder.uid_encoder.transform(user_ids)

    def user_ids(self, user_inds):
        return self.sparse_mat_builder.uid_encoder.inverse_transform(user_inds)

    def item_inds(self, item_ids):
        return self.sparse_mat_builder.iid_encoder.transform(item_ids)

    def item_ids(self, item_inds):
        return self.sparse_mat_builder.iid_encoder.inverse_transform(item_inds)

    def unknown_users_mask(self, user_ids):
        return self.sparse_mat_builder.uid_encoder.find_new_labels(user_ids)

    def unknown_items_mask(self, item_ids):
        return self.sparse_mat_builder.iid_encoder.find_new_labels(item_ids)

    def _set_data(self, train_obs, exclude_obs=None, exclude_training=True):
        """
        :param train_obs: training observations handler
        :param exclude_obs: additional observations that need to be excluded
            (e.g. observations that were not in training, but should not be predicted for test)
        :param exclude_training: whether to include training in the
            resulting excluded matrix, if True exclude_obs and train_obs are excluded
        """
        train_df = train_obs.df_obs
        self.sparse_mat_builder = train_obs.get_sparse_matrix_helper()
        self.all_users = train_df[self.sparse_mat_builder.uid_source_col].unique().astype(str)
        self.all_items = train_df[self.sparse_mat_builder.iid_source_col].unique().astype(str)
        # shuffling because np.unique() returns elements in almost sorted order by counts,
        # and it's probably not a good thing: it changes regional sparsity,
        # and at a later stage might be sampled / iterated in order
        np.random.shuffle(self.all_users)
        np.random.shuffle(self.all_items)
        self.train_mat = self.sparse_mat_builder.\
            build_sparse_interaction_matrix(train_df)
        # set training mat as default exclude_mat
        self.set_exclude_mat(exclude_obs, exclude_training=exclude_training)
        self.set_items_handler(train_obs)
        
    def set_items_handler(self, train_obs):
        if isinstance(train_obs,
                      interactions.ItemsFeaturesHandler):
            self.items_handler = interactions.ItemsFeaturesHandler(
                df_items=train_obs.df_items, item_id_col=train_obs.item_id_col)

    def set_exclude_mat(self, exclude_obs=None, exclude_training=True):
        """
        this methods allows to alter the exclusion matrix from the default of using just the training.
        An error is raised if parameters dictate no excluded data in order to prevent accidental mistake.
        :param exclude_obs: alternative / additional excluded data
        :param exclude_training: whether to exclude training data
        """
        if exclude_obs is None:
            if not exclude_training:
                raise ValueError("Got exclude_training=False, but no "
                                 "replacement exclude_obs provided. "
                                 "If you're sure you don't want to have excluded data "
                                 "at all provide an empty observation handler")
            exclude_mat = self.train_mat.copy() if self.train_mat is not None else None
        else:
            exclude_mat = self.sparse_mat_builder.\
                build_sparse_interaction_matrix(exclude_obs.df_obs)
            if exclude_training and self.train_mat is not None:
                exclude_mat += self.train_mat
        self.exclude_mat = exclude_mat

    def _reuse_data(self, other):
        self.all_users = other.all_users
        self.all_items = other.all_items
        self.sparse_mat_builder = other.sparse_mat_builder
        self.train_mat = other.train_mat
        self.exclude_mat = other.exclude_mat
        self.items_handler = other.items_handler

    def remove_unseen_users(self, user_ids, message_prefix=''):
        return self._filter_array(
            user_ids,
            encoder=self.sparse_mat_builder.uid_encoder,
            message_prefix=message_prefix,
            message_suffix='users that were not in training set.')

    def remove_unseen_items(self, item_ids, message_prefix=''):
        return self._filter_array(
            item_ids,
            encoder=self.sparse_mat_builder.iid_encoder,
            message_prefix=message_prefix,
            message_suffix='items that were not in training set.')

    @staticmethod
    def _filter_array(array, encoder, message_prefix='', message_suffix=''):
        array = np.array(array).astype(str)
        new_labels_mask = encoder.find_new_labels(array)
        n_discard = np.sum(new_labels_mask)
        if n_discard > 0:
            logger.info(
                '%s Discarding %d (out of %d) %s' %
                (message_prefix, int(n_discard), len(array), message_suffix))
        return array[~new_labels_mask]

    @staticmethod
    def _remove_self_similarities(flat_df, col1, col2):
        return flat_df[flat_df[col1].values != flat_df[col2].values].copy()

    @staticmethod
    def _eval_on_test_by_ranking_LFM(train_ranks_func, test_ranks_func,
                                     test_dfs, test_names=None, prefix='',
                                     include_train=True, k=10):
        """
        this is just to avoid the same flow twice (or more)
        :param train_ranks_func: function that return the ranks and sparse mat of training set
        :param test_ranks_func: function that return the ranks and sparse mat of a test set
        :param test_dfs: test dataframes
        :param test_names: test dataframes names
        :param prefix: prefix for this report
        :param include_train: whether to evaluate training or not
        :return: a report dataframe
        """

        # test
        if isinstance(test_dfs, pd.DataFrame):
            test_dfs = [test_dfs]

        if test_names is None or len(test_names) != len(test_dfs):
            logger.warning('No test names provided or number of names '
                           'not matching number of test DFs, using empty strings')
            test_names = [''] * len(test_dfs)

        res = []
        report_dfs = []
        full_reports = {}
        with multiprocessing.pool.ThreadPool(len(test_dfs) + int(include_train)) as pool:
            if include_train:
                res.append((prefix + 'train', pool.apply_async(train_ranks_func)))

            for test_df, test_name in zip(test_dfs, test_names):
                res.append((prefix + test_name + 'test',
                            pool.apply_async(test_ranks_func, args=(test_df,))))

            for name, r in res:
                ranks_mat, sp_mat = r.get()
                means_report, full_report = evaluation.mean_scores_report_on_ranks(
                    ranks_list=[ranks_mat], datasets=[sp_mat],
                    dataset_names=[name], k=k)
                report_dfs.append(means_report)
                full_reports.update(full_report)

        report_df = pd.concat(report_dfs, sort=False)
        return report_df, full_reports

    def get_prediction_mat_builder_adapter(
            self, mat_builder: interactions.InteractionMatrixBuilder):
        mat_builder = copy.deepcopy(mat_builder)
        mat_builder.uid_source_col = self._user_col
        mat_builder.iid_source_col = self._item_col
        mat_builder.rating_source_col = self._prediction_col
        return mat_builder

    @abc.abstractmethod
    def _get_recommendations_flat(self, user_ids, n_rec, item_ids=None,
                                  exclusions=True, **kwargs):
        return pd.DataFrame()

    @abc.abstractmethod
    def _predict_on_inds_dense(self, user_inds, item_inds):
        return np.array([])

    def get_recommendations(
            self, user_ids=None, item_ids=None, n_rec=10,
            exclusions=True, results_format='lists',
            **kwargs):

        if user_ids is not None:
            user_ids = self.remove_unseen_users(
                user_ids, message_prefix='get_recommendations: ')
        else:
            user_ids = self.all_users

        if item_ids is not None:
            item_ids = self.remove_unseen_items(
                item_ids, message_prefix='get_recommendations: ')

        recos_flat = self._get_recommendations_flat(
            user_ids=user_ids, item_ids=item_ids, n_rec=n_rec,
            exclusions=exclusions)

        if results_format == 'flat':
            return recos_flat
        else:
            return self._recos_flat_to_lists(recos_flat, n_cutoff=n_rec)

    def _check_item_ids_args(self, item_ids, target_item_ids):
        item_ids = self.remove_unseen_items(item_ids) \
            if item_ids is not None else self.all_items
        target_item_ids = self.remove_unseen_items(target_item_ids) \
            if target_item_ids is not None else None
        return item_ids, target_item_ids

    def _check_user_ids_args(self, user_ids, target_user_ids):
        user_ids = self.remove_unseen_users(user_ids) \
            if user_ids is not None else self.all_users
        target_user_ids = self.remove_unseen_users(target_user_ids) \
            if target_user_ids is not None else None
        return user_ids, target_user_ids

    def eval_on_test_by_ranking(self, test_dfs, test_names=None, prefix='rec ',
                                include_train=True, items_filter=None,
                                n_rec=100, k=10, return_full_metrics=False):

        # convert to list
        if isinstance(test_dfs, (pd.DataFrame, interactions.ObservationsDF)):
            test_dfs = [test_dfs]

        # convert to dfs
        for i in range(len(test_dfs)):
            if isinstance(test_dfs[i], interactions.ObservationsDF):
                test_dfs[i] = test_dfs[i].df_obs

        mat_builder = self.sparse_mat_builder
        pred_mat_builder = self.get_prediction_mat_builder_adapter(mat_builder)

        all_test_users = np.unique(np.concatenate([df[self._user_col].unique()
                                                   for df in test_dfs]))
        users = self.remove_unseen_users(all_test_users)

        if include_train:
            recos_flat_train = self.get_recommendations(
                user_ids=users,
                item_ids=None,
                n_rec=min(n_rec, mat_builder.n_cols),
                exclusions=False,
                results_format='flat')

        recos_flat_test = self.get_recommendations(
            user_ids=users,
            item_ids=items_filter,
            n_rec=min(n_rec, mat_builder.n_cols),
            exclusions=True,
            results_format='flat')

        ranks_all_test = pred_mat_builder.predictions_df_to_sparse_ranks(
            recos_flat_test)

        def _get_training_ranks():
            users_inds = mat_builder.uid_encoder.transform(users)
            users_inds.sort()
            # train_df = self.train_df[self.train_df[self._user_col].isin(users)].copy()
            # sp_train = self.sparse_mat_builder. \
            #     build_sparse_interaction_matrix(train_df).tocsr()
            sp_train = mat_builder.crop_rows(self.train_mat, inds_stay=users_inds)
            sp_train_ranks = pred_mat_builder. \
                filter_all_ranks_by_sparse_selection(
                sp_train, pred_mat_builder.predictions_df_to_sparse_ranks(recos_flat_train))
            return sp_train_ranks, sp_train

        def _get_test_ranks(test_df):
            sp_test = self.sparse_mat_builder. \
                build_sparse_interaction_matrix(test_df).tocsr()
            sp_test_ranks = pred_mat_builder. \
                filter_all_ranks_by_sparse_selection(sp_test, ranks_all_test)
            return sp_test_ranks, sp_test

        report_df, full_metrics = self._eval_on_test_by_ranking_LFM(
            train_ranks_func=_get_training_ranks,
            test_ranks_func=_get_test_ranks,
            test_dfs=test_dfs,
            test_names=test_names,
            prefix=prefix,
            include_train=include_train,
            k=k)

        if return_full_metrics:
            return report_df, full_metrics
        else:
            return report_df

    def get_recommendations_exact(
            self, user_ids, item_ids=None, n_rec=10,
            exclusions=True, chunksize=200, results_format='lists'):

        calc_func = functools.partial(
            self._get_recommendations_exact,
            n_rec=n_rec,
            item_ids=item_ids,
            exclusions=exclusions,
            results_format=results_format)

        chunksize = int(35000 * chunksize / self.sparse_mat_builder.n_cols)

        ret = array_math.map_batches_multiproc(calc_func, user_ids, chunksize=chunksize)
        return pd.concat(ret, axis=0, sort=False)

    def _predict_for_users_dense(self, user_ids, item_ids=None, exclusions=True):
        """
        method for calculating prediction for a grid of users and items.
        this method uses the _predict_on_inds() method that the subclasses should implement.

        :param user_ids: users ids
        :param item_ids: item ids, when None - all known items are used
        :param exclusions: when True uses default exclusion matrix sets prediction on training examples to -np.inf
        :return: a matrix of predictions (n_users, n_items)
        """

        if item_ids is None:
            item_ids = self.sparse_mat_builder.iid_encoder.classes_

        user_inds = self.user_inds(user_ids)
        item_inds = self.item_inds(item_ids)

        full_pred_mat = self._predict_on_inds_dense(user_inds, item_inds)

        if exclusions:
            exclude_mat_sp_coo = self.exclude_mat[user_inds, :][:, item_inds].tocoo()
            full_pred_mat[exclude_mat_sp_coo.row, exclude_mat_sp_coo.col] = -np.inf

        return full_pred_mat

    _predict_for_users_dense_direct = _predict_for_users_dense

    def _get_recommendations_exact(self, user_ids, item_ids=None, n_rec=10, exclusions=True,
                                   results_format='lists'):

        full_pred_mat = self._predict_for_users_dense(user_ids, item_ids,
                                                      exclusions=exclusions)

        top_inds, top_scores = array_math.top_N_sorted(full_pred_mat, n=n_rec)

        # best_ids = self.sparse_mat_builder.iid_encoder.inverse_transform(top_inds)
        best_ids = item_ids[top_inds] if item_ids is not None else \
            self.sparse_mat_builder.iid_encoder.inverse_transform(top_inds)

        return self._format_results_df(
            source_vec=user_ids, target_ids_mat=best_ids,
            scores_mat=top_scores, results_format='recommendations_' + results_format)


    def predict_for_user(self, user_id, item_ids, rank_training_last=True,
                         sort=True, combine_original_order=False):
        """
        method for predicting for one user for a small subset of items.
        optimized for minimal latency for use in real-time ranking
        will return -np.inf for combinations of unknown user / unknown items

        :param user_id: a single user ID, may be an unknown users (all predictions will be None)
        :param item_ids: a subset of item IDs, may have unknown items (prediction for those will be None)
        :param rank_training_last:  if set to True predictions for interactions seen during training
            seen during will be ranked last by being set to -np.inf
        :param combine_original_order: whether to combine predictions with original
            order of the items (if they were already ordered in a meaningful way)
        :param sort: whether to sort the result in decreasing order of prediction (best first)
        :return: a pandas DataFrame of userid | itemid | prediction,
            sorted in decresing order of prediction (if sort=True),
            with Nones for unknown users or items
        """

        user_id = str(user_id)
        item_ids = np.array(item_ids).astype(str)

        df = pd.DataFrame()
        df[self._item_col] = item_ids  # assigning first because determines length
        df[self._user_col] = user_id
        df[self._prediction_col] = -np.inf

        if np.any(self.unknown_users_mask([user_id])):
            return df

        new_items_mask = self.unknown_items_mask(item_ids)

        preds = self._predict_for_users_dense_direct(
            user_ids=[user_id],
            item_ids=item_ids[~new_items_mask],
            exclusions=rank_training_last)

        df[self._prediction_col].values[~new_items_mask] = preds.ravel()

        if combine_original_order:
            orig_score = len(item_ids) - np.arange(len(item_ids))
            preds_score = df[self._prediction_col].values
            df[self._prediction_col] = (stats.rankdata(orig_score) +
                                        stats.rankdata(preds_score)) / 2

        if sort:
            df.sort_values(self._prediction_col, ascending=False, inplace=True)

        return df[[self._user_col, self._item_col, self._prediction_col ]]  # reordering


class BasePredictorRecommender(BaseDFSparseRecommender):

    @abc.abstractmethod
    def _predict_on_inds(self, user_inds, item_inds):
        pass

    @abc.abstractmethod
    def _predict_rank(self, test_mat, train_mat=None):
        pass

    def _predict_on_inds_dense(self, user_inds, item_inds):
        n_users = len(user_inds)
        n_items = len(item_inds)
        user_inds_mat = user_inds.repeat(n_items)
        item_inds_mat = np.tile(item_inds, n_users)
        return self._predict_on_inds(user_inds_mat, item_inds_mat).reshape((n_users, n_items))

    def predict_on_df(self, df, exclusions=True, user_col=None, item_col=None):
        if user_col is not None and user_col != self.sparse_mat_builder.uid_source_col:
            df[self.sparse_mat_builder.uid_source_col] = df[user_col]
        if item_col is not None and item_col != self.sparse_mat_builder.iid_source_col:
            df[self.sparse_mat_builder.iid_source_col] = df[item_col]

        mat_builder = self.sparse_mat_builder
        df = mat_builder.remove_unseen_labels(df)
        df = mat_builder.add_encoded_cols(df)
        df[self._prediction_col] = self._predict_on_inds(
            df[mat_builder.uid_col].values, df[mat_builder.iid_col].values)

        if exclusions:
            exclude_mat_sp_coo = \
                self.exclude_mat[df[mat_builder.uid_col].values, :] \
                    [:, df[mat_builder.iid_col].values].tocoo()
            df[df[mat_builder.uid_col].isin(exclude_mat_sp_coo.row) &
               df[mat_builder.iid_col].isin(exclude_mat_sp_coo.col)][self._prediction_col] = -np.inf

        df.drop([mat_builder.uid_col, mat_builder.iid_col], axis=1, inplace=True)
        return df

    def eval_on_test_by_ranking_exact(self, test_dfs, test_names=None,
                                      prefix='lfm ', include_train=True, k=10,
                                      return_full_metrics=False):

        # convert to list
        if isinstance(test_dfs, (pd.DataFrame, interactions.ObservationsDF)):
            test_dfs = [test_dfs]

        # convert to dfs
        for i in range(len(test_dfs)):
            if isinstance(test_dfs[i], interactions.ObservationsDF):
                test_dfs[i] = test_dfs[i].df_obs

        def _get_training_ranks():
            all_test_users = np.unique(np.concatenate(
                [df[self._user_col].unique() for df in test_dfs]))
            users = self.remove_unseen_users(all_test_users)
            users_inds = self.sparse_mat_builder.uid_encoder.transform(users)
            users_inds.sort()
            sub_train_mat = self.sparse_mat_builder.crop_rows(
                self.train_mat, inds_stay=users_inds)
            ranks_mat = self._predict_rank(sub_train_mat)
            return ranks_mat, sub_train_mat

        def _get_test_ranks(test_df):
            test_sparse = self.sparse_mat_builder.\
                build_sparse_interaction_matrix(test_df)
            ranks_mat = self._predict_rank(test_sparse, self.train_mat)
            return ranks_mat, test_sparse

        report_df, full_metrics = self._eval_on_test_by_ranking_LFM(
            train_ranks_func=_get_training_ranks,
            test_ranks_func=_get_test_ranks,
            test_dfs=test_dfs,
            test_names=test_names,
            prefix=prefix,
            include_train=include_train,
            k=k)

        if return_full_metrics:
            return report_df, full_metrics
        else:
            return report_df



import multiprocessing.pool

import pandas as pd

from lightfm_pandas.utils import logging_config

N_CPUS = multiprocessing.cpu_count()

logging_config.config()


def console_settings():
    pd.set_option('display.max_colwidth', 300)
    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', 1000)


console_settings()



import os

import io
import pandas as pd
import requests
import zipfile

MOVIE_LENS_1M_URL = 'http://files.grouplens.org/datasets/movielens/ml-1m.zip'
ratings_csv_name = 'readable_ratings.csv'
users_csv_name = 'users.csv'
movies_csv_name = 'movies.csv'


def get_and_prep_data(root_dir=None):
    """
    Downloads raw data files and makes readable self explanatory csv files
    :param root_dir: choice of root data directory
    :return: paths to ratings, users, and movies csv files
    """

    if not root_dir:
        root_dir = os.path.join(os.path.abspath(os.curdir), 'out')
        os.makedirs(root_dir, exist_ok=True)

    ml_1m_dir = os.path.join(root_dir, 'ml-1m')
    rating_csv_path = os.path.join(ml_1m_dir, ratings_csv_name)
    users_csv_path = os.path.join(ml_1m_dir, users_csv_name)
    movies_csv_path = os.path.join(ml_1m_dir, movies_csv_name)

    if not os.path.exists(rating_csv_path):

        # download raw data if it's not there
        if not os.path.exists(ml_1m_dir):
            download_raw_data(root_dir)

        # make readable csv files
        prep_readable_csvs(ml_1m_dir)

    return rating_csv_path, users_csv_path, movies_csv_path


def download_raw_data(dir_path):
    r = requests.get(MOVIE_LENS_1M_URL)
    if r.ok:
        print('downloading and unzipping raw data..')
        z = zipfile.ZipFile(io.BytesIO(r.content))
        z.extractall(dir_path)
    else:
        raise ConnectionError('Failed downloading dataset from %s' % MOVIE_LENS_1M_URL)


def prep_readable_csvs(out_dir):
    # read the movies data
    movies_df = read_movies(out_dir)

    # read the users data
    users_df = pd.read_csv(os.path.join(out_dir, 'users.dat'),
                           delimiter='::', header=None,
                           names=['user_ind', 'gender', 'age', 'occupation', 'zipcode'])

    # change occupations index to occupation names
    occupations_df = get_occupation_names_df()
    users_df = pd.merge(users_df, occupations_df, left_on='occupation', right_on='index')

    # make a userid string out of all the user features: gender-age-occupation_name-zipcode
    users_df['userid'] = users_df[
        ['gender', 'age', 'occupation_name', 'zipcode']]. \
        apply(lambda x: '-'.join([str(el) for el in x]), axis=1)

    # read the ratings data
    ratings_df = pd.read_csv(os.path.join(out_dir, 'ratings.dat'),
                             delimiter='::', header=None,
                             names=['user_ind', 'item_ind', 'rating', 'timestamp'])

    # join with movie titles (itemids)
    ratings_df = pd.merge(ratings_df, movies_df[['item_ind', 'itemid']],
                          on='item_ind').drop('item_ind', axis=1)
    # join with userid strings
    ratings_df = pd.merge(ratings_df, users_df[['user_ind', 'userid']],
                          on='user_ind').drop('user_ind', axis=1)

    # remove item index from movies
    movies_df.drop('item_ind', axis=1, inplace=True)
    users_df.drop('user_ind', axis=1, inplace=True)

    # save everything
    movies_df.to_csv(os.path.join(out_dir, movies_csv_name), index=None)
    users_df.to_csv(os.path.join(out_dir, users_csv_name), index=None)
    ratings_df.to_csv(os.path.join(out_dir, ratings_csv_name), index=None)


def read_movies(out_dir):
    movies_raw = pd.read_csv(os.path.join(out_dir, 'movies.dat'),
                             delimiter='::', header=None,
                             names=['item_ind', 'title', 'genres']). \
        rename({'title': 'itemid'}, axis=1)

    # process
    genres = set()
    for l in movies_raw.genres.str.split('|').values:
        genres.update(l)

    # regex to create indicator columns
    regex = '|'.join([f"""(?P<{g.replace('-', '').replace("'", "")}>{g})""" for g in genres])

    # create indicator columns
    split = movies_raw['genres'].str.extract(regex).notnull().astype(int)

    movies_df = pd.concat([movies_raw.drop('genres', axis=1), split], axis=1, sort=False)

    return movies_df


def get_occupation_names_df():
    # form the dataset README file
    occupations = \
        """
            *  0:  "other"
            *  1:  "academic/educator"
            *  2:  "artist"
            *  3:  "clerical/admin"
            *  4:  "college/grad student"
            *  5:  "customer service"
            *  6:  "doctor/health care"
            *  7:  "executive/managerial"
            *  8:  "farmer"
            *  9:  "homemaker"
            * 10:  "K-12 student"
            * 11:  "lawyer"
            * 12:  "programmer"
            * 13:  "retired"
            * 14:  "sales/marketing"
            * 15:  "scientist"
            * 16:  "self-employed"
            * 17:  "technician/engineer"
            * 18:  "tradesman/craftsman"
            * 19:  "unemployed"
            * 20:  "writer"
        """
    occupations_df = pd.read_csv(io.BytesIO(occupations.encode()),
                                 delim_whitespace=True, header=None,
                                 names=['star', 'ind', 'occupation_name']). \
        drop(['star', 'ind'], axis=1).reset_index()
    return occupations_df



import abc
import functools
import inspect
import logging
import pickle
import threading
import time
from types import FunctionType

import psutil

from lightfm_pandas.utils import logging_config

logging_config.config()

logger = logging.getLogger(__name__)


class Config:
    """
    Set "use_instrumentation" in some common context
    in order to disable decoration logging.

    I usually leave this on all the time as the profiling is quite
    light weight when not used on very frequently running
    functions (functionality to prevent this is provided below),
    and this also serves as a nice progress indicator to
    know what long-running code is doing at any time.
    """
    use_instrumentation = True
    min_time_seconds = 1
    level = logging.INFO


def variable_info(result):
    """
    Create a string that gives some useful information about an output of a function.
    :param result: anything
    :return: string with some useful information about the result.
    """
    if hasattr(result, 'shape'):
        shape_str = 'shape: %s' % str(result.shape)
    elif isinstance(result, tuple) and len(result) <= 3:
        shape_str = 'tuple: (' + ','.join([variable_info(el) for el in result]) + ')'
    elif hasattr(result, '__len__'):
        shape_str = 'len: %s' % str(len(result))
    elif hasattr(result, '__dict__'):
        shape_str = 'keys: %s' % str(len(result.__dict__.keys()))
    else:
        result_str = str(result)
        shape_str = result_str if len(result_str) <= 50 else (str(result)[:50] + '...')

    return str(type(result)) + ', ' + shape_str


def log_time_and_shape(fn):
    """
    Decorates the function to log it's run time, resource usage (CPU and memory),
    and output shape / info.
    The logging statements are indented according to the stack depth in order to have a
    visual representation of depth.
    :param fn: function to be decorated.
    :return: decorated function.
    """

    @functools.wraps(fn)
    def inner(*args, **kwargs):
        with ResourceMonitor() as monitor:
            result = fn(*args, **kwargs)

        if monitor.elapsed >= Config.min_time_seconds:
            msg = (' ' * get_stack_depth() +
                   f'{function_name_with_class(fn)}, elapsed: {monitor.elapsed:.2f} sec, '
                   f'returned: {variable_info(result)}, '
                   f'sys mem: {monitor.current_memory}%'
                   f'(peak:{monitor.peak_memory}%) '
                   f'cpu:{int(monitor.avg_cpu_load)}%'
                   )

            logger.log(Config.level, msg)

        return result

    return inner if Config.use_instrumentation else fn


class ResourceMonitor:
    """
    Class for monitoring the CPU and memory of a function call.
    """

    def __init__(self, interval=0.2):
        self.interval = interval
        self._init_counters()
        self._thread = None
        self._run_condition = False
        self.elapsed = None

    def _init_counters(self):
        self._start_time = time.time()
        self.current_memory = 0
        self.peak_memory = 0
        self.avg_cpu_load = 0
        self._n_measurements = 0

    def __del__(self):
        self.stop()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()

    @staticmethod
    def _current():
        try:
            return psutil.virtual_memory().percent, psutil.cpu_percent()
        except KeyError:
            # for some reason there's a KeyError: ('psutil',) in psutil
            return 0, 0
        except Exception as e:
            logger.exception(e)
            return 0, 0

    def _measure(self):
        cur_mem, cur_cpu = self._current()
        self.current_memory = cur_mem
        self.peak_memory = max(self.peak_memory, cur_mem)
        self.avg_cpu_load = (self.avg_cpu_load * self._n_measurements + cur_cpu) / \
                            (self._n_measurements + 1)
        self._n_measurements += 1

    def _thread_loop(self):
        while self._run_condition:
            self._measure()
            time.sleep(self.interval)

    def start(self):
        self._init_counters()
        if self._thread is not None:
            self._thread.join(0)
        self._run_condition = True
        self._thread = threading.Thread(target=self._thread_loop, name='ResourceMonitor')
        self._thread.start()
        return self

    def stop(self):
        self.elapsed = time.time() - self._start_time
        self._run_condition = False
        self._measure()


def get_stack_depth():
    try:
        return len(inspect.stack(context=0))
    except (IndexError, RuntimeError) as e:
        # there is a bug in inspect module: https://github.com/ipython/ipython/issues/1456/
        # another one: https://bugs.python.org/issue13487
        return 0


def function_name_with_class(fn):
    cls = get_class_that_defined_method(fn)
    cls_str = cls.__name__ + '.' if cls else ''
    return cls_str + fn.__name__


def get_class_that_defined_method(meth):
    # from https://stackoverflow.com/questions/3589311/
    # get-defining-class-of-unbound-method-object-in-python-3/25959545#25959545
    # modified to return first parent in reverse MRO
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__)[::-1]:
            if cls.__dict__.get(meth.__name__) is meth:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def decorate_all_metaclass(decorator):
    """
    Decorate all methods and classmethods (unless excluded) with the same decorator.

    altered from https://stackoverflow.com/questions/10067262/automatically-decorating-every-instance-method-in-a-class
    """

    # check if an object should be decorated
    def do_decorate(attr, value):
        return ('__' not in attr and
                isinstance(value, (FunctionType, classmethod)) and
                getattr(value, 'decorate', True))

    class DecorateAll(abc.ABCMeta):
        def __new__(cls, name, bases, dct):
            if dct.get('decorate', True):
                for attr, value in dct.items():
                    if do_decorate(attr, value):
                        if isinstance(value, classmethod):
                            dct[attr] = classmethod(decorator(value.__func__))
                        else:
                            dct[attr] = decorator(value)
            return super(DecorateAll, cls).__new__(cls, name, bases, dct)

        def __setattr__(self, attr, value):
            if do_decorate(attr, value):
                value = decorator(value)
            super(DecorateAll, self).__setattr__(attr, value)

    return DecorateAll


class LogLongCallsMeta(metaclass=decorate_all_metaclass(log_time_and_shape)):
    """
        Class to provide automatic logging decoration of all methods for extending classes.

        Note on overhead for frequently invoked methods:
            If your class includes methods that are invoked very frequently,
            e.g. thousands / millions of times per second, you should use `do_not_decorate`
            decorator in order to prevent the logging decorator being applied to those methods,
            since the overhead of monitoring will be substantial in those cases.
        """

    @property
    def logging_decorator(self):
        """
        this is for decorating inner scope functions
        :return: the logging decorator if verbose is True, empty decorator otherwise
        """
        return log_time_and_shape

    @staticmethod
    def do_not_decorate(f):
        """
        this is for excluding functions from decorating
        :return: the logging decorator if verbose is True, empty decorator otherwise
        """
        setattr(f, 'decorate', False)
        return f


def log_errors(message=None, return_on_error=None):
    def decorator(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            nonlocal message, return_on_error
            try:
                return fn(*args, **kwargs)
            except Exception as e:
                logger.exception(e)
                fn_str = function_name_with_class(fn)
                msg_str = ', Message: %s' % message if message else ''
                logger.error('Failed: %s, Error: %s %s' % (fn_str, str(e), msg_str))
                return return_on_error

        return inner

    return decorator


def pickle_size_mb(obj):
    return pickle.dumps(obj).__sizeof__() / 1e6



import importlib

import numpy as np
import pandas as pd
import pandas.core.algorithms
import pandas.core.dtypes.common
import sklearn.utils
from sklearn import preprocessing


class PDLabelEncoder(preprocessing.LabelEncoder):
    """
    adapted and sped up further from here: https://github.com/scikit-learn/scikit-learn/issues/7432
    a blazingly fast version of encoder that's using Pandas encoding
    which is using hash tables instead of sorted arrays
    """

    pd_cat_module = importlib.import_module(pd.Categorical.__module__)

    def fit(self, y):
        y = sklearn.utils.column_or_1d(y, warn=True)
        _, self.classes_ = pd.factorize(y, sort=True)
        self._cat_dtype = self.pd_cat_module.CategoricalDtype(self.classes_)
        self._dtype = self._cat_dtype.categories.dtype
        self._table = self._get_table_for_categories(self.classes_, self._cat_dtype.categories)
        return self

    @classmethod
    def _get_table_for_categories(cls, values, categories):
        # ripped out of _get_codes_for_values() in pandas Categorical module
        if not cls.pd_cat_module.is_dtype_equal(values.dtype, categories.dtype):
            values = pandas.core.dtypes.common.ensure_object(values)
            categories = pandas.core.dtypes.common.ensure_object(categories)

        hash_klass, vals = pandas.core.algorithms._get_data_algo(values)
        _, cats = pandas.core.algorithms._get_data_algo(categories)
        t = hash_klass(len(cats))
        t.map_locations(cats)
        return t

    def check_is_fitted(self):
        sklearn.utils.validation.check_is_fitted(
            self, ['classes_', '_cat_dtype', '_table', '_dtype'])

    def transform(self, y, check_labels=True):
        self.check_is_fitted()
        y = sklearn.utils.column_or_1d(y, warn=True)

        ## slower because it creates the hash table every time
        # trans_y = pd.Categorical(y, dtype=self._cat_dtype).codes.copy()
        trans_y = self.pd_cat_module.coerce_indexer_dtype(
            indexer=self._table.lookup(y.astype(self._dtype)),
            categories=self._cat_dtype.categories)

        if check_labels:
            mask_new_labels = self._new_labels_locs(trans_y)
            if np.any(mask_new_labels):
                raise ValueError("y contains new labels: %s" %
                                 str(np.unique(y[mask_new_labels])))

        return trans_y

    def inverse_transform(self, y):
        y = np.array(y)
        shape = y.shape
        return super().inverse_transform(y.ravel()).reshape(shape)

    def _new_labels_locs(self, trans_y):
        return trans_y == -1

    def find_new_labels(self, y):
        return self._new_labels_locs(self.transform(y, check_labels=False))

    def __getstate__(self):
        # custom pickling behaviour because the table object is unpickleable
        state = super().__getstate__()
        state['_table'] = None
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self._table = self._get_table_for_categories(self.classes_, self._cat_dtype.categories)



from setuptools import setup, find_packages
from os import path

version = '0.2.1'

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='lightfm_pandas',
    version=version,
    description='LightFM convenience tools',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/artdgn/lightfm-pandas',
    author='Arthur Deygin',
    author_email='arthurdgn@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='recommendations machine learning',
    packages=find_packages(exclude=['tests', 'examples']),
    install_requires=open('requirements.in').read().split('\n'),
)



import abc

import pandas as pd
import numpy as np
import scipy.sparse as sp

from lightfm_pandas.modeling import base
from lightfm_pandas.utils import array_math


class BaseFactorizationRecommender(base.BasePredictorRecommender):

    @abc.abstractmethod
    def _get_item_factors(self, mode=None):
        return  np.array([0]), np.array([0])

    @abc.abstractmethod
    def _get_user_factors(self, mode=None):
        return np.array([0]), np.array([0])

    def _factors_to_dataframe(self, factor_func, include_biases=False):
        b, f = factor_func()
        return pd.DataFrame(
            index=np.arange(f.shape[0]),
            data=np.concatenate([b, f], axis=1) if include_biases else f)

    def user_factors_dataframe(self, include_biases=False):
        return self._factors_to_dataframe(self._get_user_factors, include_biases=include_biases)

    def item_factors_dataframe(self, include_biases=False):
        return self._factors_to_dataframe(self._get_item_factors, include_biases=include_biases)

    def get_similar_items(self, item_ids=None, target_item_ids=None, n_simil=10,
                          remove_self=True, embeddings_mode=None,
                          simil_mode='cosine', results_format='lists'):
        """
        uses learned embeddings to get N most similar items

        :param item_ids: vector of item IDs
        :param n_simil: number of most similar items to retrieve
        :param remove_self: whether to remove the the query items from the lists (similarity to self should be maximal)
        :param embeddings_mode: the item representations to use for calculation:
             None (default) - means full representations
             'external_features' - calculation based only external features (assumes those exist)
             'no_features' - calculation based only on internal features (assumed identity mat was part of the features)
        :param simil_mode: mode of similairyt calculation:
            'cosine' (default) - cosine similarity bewtween representations (normalized dot product with no biases)
            'dot' - unnormalized dot product with addition of biases
            'euclidean' - inverse of euclidean distance
        :param results_format:
            'flat' for dataframe of triplets (source_item, similar_item, similarity)
            'lists' for dataframe of lists (source_item, list of similar items, list of similarity scores)

        :return: a matrix of most similar IDs [n_ids, N], a matrix of score of those similarities [n_ids, N]
        """

        item_ids, target_item_ids = self._check_item_ids_args(item_ids, target_item_ids)

        biases, representations = self._get_item_factors(mode=embeddings_mode)

        best_ids, best_scores = array_math.most_similar(
            source_ids=item_ids,
            target_ids=target_item_ids,
            source_encoder=self.sparse_mat_builder.iid_encoder,
            source_mat=representations,
            source_biases=biases,
            n=n_simil+1 if remove_self else n_simil,
            simil_mode=simil_mode,
        )

        simil_df = self._format_results_df(
            item_ids, target_ids_mat=best_ids,
            scores_mat=best_scores, results_format='similarities_flat')

        if remove_self:
            simil_df = self._remove_self_similarities(
                simil_df, col1=self._item_col_simil, col2=self._item_col)

        if 'lists' in results_format:
            simil_df = self._simil_flat_to_lists(simil_df, n_cutoff=n_simil)

        return simil_df

    def get_similar_users(self, user_ids=None, target_user_ids=None, n_simil=10, remove_self=True,
                          simil_mode='cosine'):
        """
        same as get_similar_items but for users
        """
        user_ids, target_user_ids = self._check_user_ids_args(user_ids, target_user_ids)

        user_biases, user_representations = self._get_user_factors()

        best_ids, best_scores = array_math.most_similar(
            source_ids=user_ids,
            target_ids=target_user_ids,
            source_encoder=self.sparse_mat_builder.uid_encoder,
            source_mat=user_representations,
            source_biases=user_biases,
            n=n_simil,
            simil_mode=simil_mode,
        )

        simil_df = self._format_results_df(
            user_ids, target_ids_mat=best_ids,
            scores_mat=best_scores, results_format='similarities_flat'). \
            rename({self._item_col_simil: self._user_col}, axis=1)
        # this is UGLY, if this function is ever useful, fix this please (the renaming shortcut)

        if remove_self:
            simil_df = self._remove_self_similarities(
                simil_df, col1=self._user_col, col2=self._item_col)

        simil_df = self._recos_flat_to_lists(simil_df, n_cutoff=n_simil)

        return simil_df

    def _get_recommendations_flat(
            self, user_ids, n_rec, item_ids=None, exclusions=True,
            item_features_mode=None, use_biases=True):

        user_biases, user_representations = self._get_user_factors()
        item_biases, item_representations = self._get_item_factors(mode=item_features_mode)

        if not use_biases:
            user_biases, item_biases = None, None

        best_ids, best_scores = array_math.most_similar(
            source_ids=user_ids,
            target_ids=item_ids,
            source_encoder=self.sparse_mat_builder.uid_encoder,
            target_encoder=self.sparse_mat_builder.iid_encoder,
            source_mat=user_representations,
            target_mat=item_representations,
            source_biases=user_biases,
            target_biases=item_biases,
            exclude_mat_sp=self.exclude_mat if exclusions else None,
            n=n_rec,
            simil_mode='dot',
        )

        return self._format_results_df(
            source_vec=user_ids, target_ids_mat=best_ids, scores_mat=best_scores,
            results_format='recommendations_flat')

    def _predict_for_users_dense_direct(self, user_ids, item_ids=None, exclusions=True):
        """
        method for calculating prediction for a grid of users and items
        directly from the calculated factors. this method is faster for smaller inputs
        an can be further sped up by employing batched multiprocessing (as
        used in similarity / recommendation calculations)

        :param user_ids: users ids
        :param item_ids: item ids, when None - all known items are used
        :param exclusions: when True sets prediction on excluded examples to -np.inf
        :return: a matrix of predictions (n_users, n_items)
        """
        if item_ids is None:
            item_ids = self.sparse_mat_builder.iid_encoder.classes_

        user_inds = self.user_inds(user_ids)
        item_inds = self.item_inds(item_ids)

        user_biases, user_factors = self._get_user_factors()
        item_biases, item_factors = self._get_item_factors()

        scores = np.dot(user_factors[user_inds, :], item_factors[item_inds, :].T)

        if user_biases is not None:
            scores = (scores.T + user_biases[user_inds]).T

        if item_biases is not None:
            scores += item_biases[item_inds]

        if sp.issparse(scores):
            scores = scores.toarray()
        else:
            scores = np.array(scores)

        full_pred_mat = scores

        if exclusions:
            exclude_mat_sp_coo = self.exclude_mat[user_inds, :][:, item_inds].tocoo()
            full_pred_mat[exclude_mat_sp_coo.row, exclude_mat_sp_coo.col] = -np.inf

        return full_pred_mat

    def _predict_for_items_dense_direct(self, items_ids_source, item_ids_target=None):
        """ dense similarity predictions from items to items """
        item_inds_s = self.item_inds(items_ids_source)
        item_inds_t = self.item_inds(item_ids_target)

        item_biases, item_factors = self._get_item_factors()

        scores = np.dot(item_factors[item_inds_s, :], item_factors[item_inds_t, :].T)

        if item_biases is not None:
            scores = (scores.T + item_biases[item_inds_s]).T
            scores += item_biases[item_inds_t]

        if sp.issparse(scores):
            scores = scores.toarray()
        else:
            scores = np.array(scores)
        return scores



